package iHMMuneAlign;

import java.io.File;
import java.io.PrintStream;
import java.io.*;
import java.util.Vector;
//import javax.swing.JTextArea;
import org.biojava.bio.Annotation;
import org.biojava.bio.BioException;
import org.biojava.bio.dp.*;
import org.biojava.bio.seq.DNATools;
import org.biojava.bio.seq.*;
import org.biojava.bio.symbol.*;

// Referenced classes of package iHMMuneAlign:
//            GlobalDefines, HTMLForEXCELWriter, HTMLforWWW_Writer, BestVGeneFinder, 
//            PostAlignmentResult, A_Score, TrailingJGeneVectorFinder, VpnpDpnpJCnoC, 
//            StateInfo, PostAlignProcessing, GeneInfo, Codon, 
//            JGeneReadingFrame, HTML_Writer, RegionInfo, ProbabilityHolder, 
//            MutabilityScore

class AlignmentThread extends Thread
    implements GlobalDefines
{

    Sequence sequence;
    String txtFileName;
    String htmlFileName;
    ProbabilityHolder probHolder;
    MutabilityScore mutabilityScore;
    File V_GENES_FILE;
    File D_GENES_FILE;
    File J_GENES_FILE;
    byte alignmentType;
    int dGeneAcceptanceType;
    HTML_Writer htmlWriter;
    String scoringMatrix;
    int minJLength;
    int alignmentModelType;
    int numOfVgenes;

    public AlignmentThread(Sequence sequence, File V_GENES_FILE, File D_GENES_FILE, File J_GENES_FILE, String htmlFileName, ProbabilityHolder probHolder,
            MutabilityScore mutabilityScore, char html_writer_type, byte alignmentType, int dGeneAcceptanceType, String scoringMatrix, int minJLength, int alignmentModelType, int numOfVgenes)
    {
        htmlWriter = null;
        this.sequence = sequence;
        //this.out = out;
        this.htmlFileName = htmlFileName;
        this.probHolder = probHolder;
        this.mutabilityScore = mutabilityScore;
        this.V_GENES_FILE = V_GENES_FILE;
        this.D_GENES_FILE = D_GENES_FILE;
        this.J_GENES_FILE = J_GENES_FILE;
        this.alignmentType = alignmentType;
        this.dGeneAcceptanceType = dGeneAcceptanceType;
        this.scoringMatrix = scoringMatrix;
        this.minJLength = minJLength;
        this.alignmentModelType = alignmentModelType;
        this.numOfVgenes = numOfVgenes;
        htmlWriter = getHtmlWriter(html_writer_type);
    }

    private HTML_Writer getHtmlWriter(char html_writer_type)
    {
        switch(html_writer_type)
        {
        case 1: // '\001'
            return new HTMLForEXCELWriter();

        case 2: // '\002'
            return new HTMLforWWW_Writer();

        case 3: // '\003'
            return new HTMLforCSV_Writer();
        }

        throw new Error("No HTML Writer Type was selected");
    }

    private Object startAlignment()
    {
        String sequenceString = removeFastaStyleWhiteSpace(sequence.seqString());
        String sequenceName = sequence.getName();
        String errorType = "";
        System.out.println("about to search for the v genes");
        BestVGeneFinder vfinder = new BestVGeneFinder(V_GENES_FILE);
        int MIN_UMS_ALIGNMENT_END_OFFSET = 12;
        int MIN_IGHJ_ALIGNMENT_START_OFFSET = minJLength;
        //set the MaxScore for the identification of the first Vgene to the largest possible number as we just want to
        float prevMaxScore = java.lang.Float.MAX_VALUE;
        PostAlignmentResult alignment_results = null;
        
        //if we want to return the results for more than one Vgene do the processing here...
        //have already processed the result for a single V so only need to do this is the numOfVgenes > 1
        int vIndex;
        for (vIndex=0; vIndex<numOfVgenes; vIndex++)
        {

            //find the max scoring alignment if we are getting any subsequent v genes
            if (vIndex > 0) {
               // if previous alignment successful then get the best score for the last alignment
               prevMaxScore = alignment_results.getAlignmentScore();
            }


            //changed to create the object earlier
            alignment_results = vfinder.getResult(sequenceString, MIN_UMS_ALIGNMENT_END_OFFSET, scoringMatrix, prevMaxScore);
            //PostAlignmentResult alignment_results = vfinder.getResult(sequenceString, MIN_UMS_ALIGNMENT_END_OFFSET, scoringMatrix, prevMaxScore);
            //PostAlignmentResult alignment_results = vfinder.getResult(sequenceString);
            
            //get gaps in the best alignment
            int alignment_gap_no = 0;
            alignment_gap_no = alignment_results.getAlignmentGaps();

            //if there are gaps print msg and skip the hmm model creation as gaps not yet implemented
            if(alignment_gap_no > 0)
            {
                 //throw new Error("VGene finder result equal NULL");

                 //do something more meaningful then just throwing an error
                 System.out.println("There were gaps in the IGHV gene. iHMMune-align cannot proceed with HMM creation.");

                 //get the IGHV name
                 String VGene_name = alignment_results.getVGeneName();
                 //get the input seq that matches to the IGHV region
                 char[] alignedInput = alignment_results.getAlignedInput();
                 //convert the alignedInput to a string
                 String vOutput = new String(alignedInput);
                 //convert the output to lower case -- note: not showing any match/mismatch details just where the gaps are
                 vOutput = vOutput.toLowerCase();

                 //set the error to return
                 //errorType = "NA - Gaps in IGHV gene";
                 errorType = "NA - Gaps in IGHV gene " + VGene_name + " with " + alignment_gap_no + " gaps";

                 //call the html writer to create the output file - or to append to the output file if already exists and alignment type =4
                 htmlWriter.createErrDocument(htmlFileName ,sequenceName, alignmentType, errorType, vOutput);

                 // close the file once we are done writing the output
                 if(htmlWriter != null)
                 {
                    htmlWriter.closeHTMLFile();
                 }
        
                 //return
                 //return null;   // this will exit the loop and skip subsequent IGHV genes (if there are any remaining to be sought)

            } else { // not gaps in alignment so keep processing
                String from_alignment_start_UMS_string = alignment_results.getUMSfromAlignmentString();
                String from_alignment_start_VGene_string = alignment_results.getVGeneAlignmentString();
                String UMS_name = alignment_results.getUMSname();
                String VGene_name = alignment_results.getVGeneName();
                Sequence from_alignment_start_UMS_seq = null;
                Sequence from_alignment_start_VGene_seq = null;
                try
                {
                    from_alignment_start_UMS_seq = DNATools.createDNASequence(from_alignment_start_UMS_string, UMS_name);
                    from_alignment_start_VGene_seq = DNATools.createDNASequence(from_alignment_start_VGene_string, VGene_name);
                }
                catch(IllegalSymbolException illse)
                {
                    throw new Error(illse.getMessage());
                }

                int VGene_start_offset = alignment_results.getColOffset();
                int completeVGeneLength = VGene_start_offset + from_alignment_start_VGene_string.length();
                double A_probability = A_Score.A_probability(from_alignment_start_VGene_string, from_alignment_start_UMS_string, VGene_start_offset, null);
                System.out.println("A_probability = " + A_probability);
                //if the V was too short for A_prob to be calculated then just output
                if (A_probability > 1) {
                  //call the functions to output results
                  if (A_probability == 2) {
                     System.out.println("Invalid A_probability due to short IGHV gene");
                  }
                  if (A_probability == 3) {
                    System.out.println("Invalid A_probability due to too great truncation of the 5` end of the IGHV");
                  }
                  if (A_probability == 4) {
                    System.out.println("Invalid A_probability due to differing lengths of the IGHV and corresponding input sequence");
                  }
        
                 errorType = "NA - IGHV too short";
                 String vOutput = "NA";
        
                 //call the html writer to create the output file - or to append to the output file if already exists and alignment type =4
                 htmlWriter.createErrDocument(htmlFileName ,sequenceName, alignmentType, errorType, vOutput);
        
                 // close the file once we are done writing the output
                 if(htmlWriter != null)
                 {
                    htmlWriter.closeHTMLFile();
                 }
        
                 //return
                 return null;

                } else { //we have an A_prob so continue with the model construction
                    //determine if there is any sequence downstream of the IGHJ gene
                    //TrailingJGeneVectorFinder vectorFinder = new TrailingJGeneVectorFinder(); //changed to actually pass the input j repertoire
                    TrailingJGeneVectorFinder vectorFinder = new TrailingJGeneVectorFinder(J_GENES_FILE);
                    String UMS_no_C = vectorFinder.getResult(from_alignment_start_UMS_string, MIN_IGHJ_ALIGNMENT_START_OFFSET, scoringMatrix);
      
                    //need to check that the TrailingJGeneVector finder actually found a suitable IGHJ match
                    if (UMS_no_C == "NA")  {
                        //the min length for the J gene was not satisfied
            
                        //want to stop further processing and just give some meaningful output
                        System.out.println("IGHJ gene could not be found of a suitable length. Min length specified as " + MIN_IGHJ_ALIGNMENT_START_OFFSET);
            
                        errorType = "NA - IGHJ too short";
                        String vOutput = "NA";

                        //call the html writer to create the output file - or to append to the output file if already exists and alignment type =4
                        htmlWriter.createErrDocument(htmlFileName ,sequenceName, alignmentType, errorType, vOutput);
            
                        // close the file once we are done writing the output
                        if(htmlWriter != null)
                        {
                           htmlWriter.closeHTMLFile();
                        }
            
                         //return
                         return null;
                      }
    
                  String from_alignment_start_UMS_no_C_string = null;
                  Sequence from_alignment_start_UMS_no_C_seq = null;
                  org.biojava.bio.dp.MarkovModel markov_model;
                  if(UMS_no_C != null)
                  {
                      if (UMS_no_C == "NA") {
                        //we have no match to the IGHJ seq, but we are allowing partial seqs
                         VpnpDpnpJCnoC vdj = new VpnpDpnpJCnoC(mutabilityScore);
                         markov_model = vdj.createModel(from_alignment_start_VGene_seq, VGene_start_offset, completeVGeneLength, D_GENES_FILE, J_GENES_FILE, probHolder, false, A_probability, alignmentModelType);
        
                      } else {
                        from_alignment_start_UMS_no_C_string = UMS_no_C;
                        try
                        {
                            from_alignment_start_UMS_no_C_seq = DNATools.createDNASequence(UMS_no_C, "UMS_name");
                        }
                        catch(IllegalSymbolException illse)
                        {
                            throw new Error(illse.getMessage());
                        }
                        VpnpDpnpJCnoC vdj = new VpnpDpnpJCnoC(mutabilityScore);
                        markov_model = vdj.createModel(from_alignment_start_VGene_seq, VGene_start_offset, completeVGeneLength, D_GENES_FILE, J_GENES_FILE, probHolder, true, A_probability, alignmentModelType);
                      }
                  } else
                  {
                      //means that the IGHJ length found is equal to or less than full length germline gene
                      //possible that there is truncation of the IGHJ
                      System.out.println("NO C_Region Removal");
                      from_alignment_start_UMS_no_C_string = from_alignment_start_UMS_string;
                      from_alignment_start_UMS_no_C_seq = from_alignment_start_UMS_seq;
                      VpnpDpnpJCnoC vdj = new VpnpDpnpJCnoC(mutabilityScore);
                      markov_model = vdj.createModel(from_alignment_start_VGene_seq, VGene_start_offset, completeVGeneLength, D_GENES_FILE, J_GENES_FILE, probHolder, false, A_probability, alignmentModelType);
                  }

                  System.out.println("about to create DP");
                  DP dp = null;
                  try
                  {
                      dp = DPFactory.DEFAULT.createDP(markov_model);
                  }
                  catch(BioException bioe)
                  {
                      throw new Error(bioe.getMessage());
                  }
                  System.out.println("Dp created");
    
                  SymbolList res_array[] = {
                      from_alignment_start_UMS_no_C_seq
                  };
                  if(res_array == null)
                  {
                      System.out.println("res_array == null");
                      System.exit(0);
                  }
    
                  System.out.println("about to create state path viterbi");
                  StatePath v = null;
                  try
                  {
                      v = dp.viterbi(res_array, ScoreType.PROBABILITY);
                  }
                  catch(IllegalAlphabetException illalphe)
                  {
                      throw new Error(illalphe.getMessage());
                  }
                  catch(IllegalSymbolException illsymbole)
                  {
                      throw new Error(illsymbole.getMessage());
                  }
                  catch(IllegalTransitionException illsymbole)
                  {
                      throw new Error(illsymbole.getMessage());
                  }
                  catch(BioException bioex) {
                    throw new Error(bioex.getMessage());
                  }
                  //catch(Exception ex)
                  //{
                  //  throw new Error(ex.getMessage());
                  //}
    
                  double statePathProbability = v.getScore();
                  System.out.println("state path viterbi created");
                  SymbolList viterbi_state_seq = v.symbolListForLabel(StatePath.STATES);
                  
                  
                  //output
                  //create a file to store the output in --> this is separate to the selected HTML/CSV/WWW type of output
                  //PrintWriter fostream = null;
                  //String outputStreamFile = htmlFileName + "fostream";
                  
                  //fostream = new PrintWriter (new FileWriter (outputStreamFile, true));
                  
                  //generate the output
                  displayStateInfo(viterbi_state_seq, from_alignment_start_UMS_no_C_string, sequenceName);
                  displayEmissionInfo(viterbi_state_seq, from_alignment_start_UMS_no_C_string, sequenceName, statePathProbability, htmlFileName);
        
                  //attempt to invoke garbage collection to deal with mem usages
                  System.gc();
                  System.out.println("HINTING TO Garbage Collection");


              } //end of else statement for V with A_prob

          } //end else for if IGHV alignment was NULL
          
       }//end of processing for extra IGHV gene options

       //done with the alignments....
       return null;

    }

    private void displayStateInfo(SymbolList viterbi_state_seq, String UMS, String selected_ums_name)
    {
        String stateResult = "";
        String umsResult = "";
        String preEmitResult = "";
        String matchResult = "";
        System.out.println("Pre Emission");
        //fostream.println("Pre Emission");
        int ums_index = 0;
        for(int i = 1; i <= viterbi_state_seq.length(); i++)
        {
            Annotation annotation = viterbi_state_seq.symbolAt(i).getAnnotation();
            StateInfo si = (StateInfo)annotation.getProperty(null);
            char geneToken = si.geneToken;
            char statePreEmit = si.preEmissionSymbol;
            if(geneToken != 'X')
            {
                if(geneToken == '2' || geneToken == '1')
                {
                    stateResult = stateResult + 'N';
                    preEmitResult = preEmitResult + statePreEmit;
                } else
                if(geneToken == '3' || geneToken == '4' || geneToken == '5' || geneToken == '6')
                {
                    stateResult = stateResult + 'P';
                    preEmitResult = preEmitResult + statePreEmit;
                } else
                {
                    stateResult = stateResult + geneToken;
                    preEmitResult = preEmitResult + statePreEmit;
                }
            }
            System.out.print(statePreEmit);
            if(geneToken != 'X')
            {
                char UMSEmit = UMS.charAt(ums_index++);
                umsResult = umsResult + UMSEmit;
                if(statePreEmit == '?')
                {
                    matchResult = matchResult + '?';
                } else
                if(statePreEmit == UMSEmit)
                {
                    matchResult = matchResult + '.';
                } else
                {
                    matchResult = matchResult + '|';
                }
            }
        }

        //System.out.println("******** Result of aligning " + selected_ums_name + "  *******" + "\n");
        //System.out.println("\n");
        //System.out.println("emitting state : " + stateResult + "\n");
        //System.out.println("ums region     : " + umsResult + "\n");
        //System.out.println("nucl. in gene  : " + preEmitResult + "\n");
        //System.out.println("match result   : " + matchResult + "\n");
        
        //fostream.println("******** Result of aligning " + selected_ums_name + "  *******" + "\n");
        //fostream.out.println("\n");
        //fostream.out.println("emitting state : " + stateResult + "\n");
        //fostream.out.println("ums region     : " + umsResult + "\n");
        //fostream.out.println("nucl. in gene  : " + preEmitResult + "\n");
        //fostream.out.println("match result   : " + matchResult + "\n");
    }

    private void displayEmissionInfo(SymbolList viterbi_state_seq, String UMS, String selected_ums_name, double statePathProbability, String htmlFileName)
    {
        int state_index = 1;
        int UMS_index = 0;
        PrintWriter stream = null;
        try {
	        if (stream == null) {
	        	String outputFileName = htmlFileName + ".fostream";
	        	boolean append = true;
	        	stream = new PrintWriter (new FileWriter (outputFileName, append));
	        }
        }
        catch(Exception ex)
        {
            System.out.println("couldn't open new file: " + ex.getMessage());
        }
        
        
        //System.out.println("\n*** Details ***\n");
        //System.out.println("\n");
        //System.out.println("State Path Probability = " + statePathProbability);
        //System.out.println("\n");
        //System.out.println("\n");
        stream.println("*** Details ***");
        //stream.println("\n");
        stream.println("State Path Probability = " + statePathProbability);
        //stream.println("\n");
        //stream.println("\n");
        
        GeneInfo VGeneInfo = displayGene(UMS, viterbi_state_seq, 'V', stream);
        RegionInfo p1Region = displayNRegion(UMS, viterbi_state_seq, '3', stream);
        RegionInfo n1Region = displayNRegion(UMS, viterbi_state_seq, '1', stream );
        RegionInfo p2Region = displayNRegion(UMS, viterbi_state_seq, '4', stream);
        GeneInfo DGeneInfo = displayGene(UMS, viterbi_state_seq, 'D', stream);
        RegionInfo p3Region = displayNRegion(UMS, viterbi_state_seq, '5', stream);
        RegionInfo n2Region = displayNRegion(UMS, viterbi_state_seq, '2', stream);
        RegionInfo p4Region = displayNRegion(UMS, viterbi_state_seq, '6', stream);
        GeneInfo JGeneInfo = displayGene(UMS, viterbi_state_seq, 'J', stream);
        PostAlignProcessing postAlign = new PostAlignProcessing();
        boolean acceptedDGene = postAlign.acceptDGene(DGeneInfo, dGeneAcceptanceType, n1Region, n2Region);
        if(!acceptedDGene)
        {
            DGeneInfo.acceptedAlignment = false;
        }
        Vector stopCodons = postAlign.noStopCodons(VGeneInfo, UMS);
        if(stopCodons.size() == 0)
        {
            boolean sequenceHasStopCodons = false;
            //System.out.println("NO STOP CODONS");
            stream.println("NO STOP CODONS");
        } else
        {
            boolean sequenceHasStopCodons = true;
            //System.out.println("/n****/n*****/n**** stop codons\n*****\n");
            stream.println("**** stop codons *****");
            Codon stopCodon = (Codon)stopCodons.get(0);
            //System.out.println("at pos: " + stopCodon.nuclSeqPos);
            //System.out.println("codon: " + new String(stopCodon.codon));
            stream.println("at pos: " + stopCodon.nuclSeqPos);
            stream.println("codon: " + new String(stopCodon.codon));
        }
        JGeneReadingFrame jGeneReadingFrame = new JGeneReadingFrame(UMS, JGeneInfo, VGeneInfo);
        boolean isJGeneInFrame = jGeneReadingFrame.isJGeneInFrame();
        int relativeMotifPosition = jGeneReadingFrame.getWGXGmotifInAlignedJGenePos();
        
        htmlWriter.createDocument(htmlFileName, selected_ums_name, VGeneInfo, p1Region, n1Region, p2Region, DGeneInfo, p3Region, n2Region, p4Region, JGeneInfo, selected_ums_name, UMS, statePathProbability, alignmentType, stopCodons, relativeMotifPosition, isJGeneInFrame);
        if(htmlWriter != null)
        {
            htmlWriter.closeHTMLFile();
        }
        
        if(stream != null)
        {        
        	System.out.println("Mutational File Output Closed for editing");
        	stream.close();
        	stream = null;
        	
        }
    }

    private RegionInfo displayNRegion(String ums, SymbolList viterbi_state_seq, char state_token, PrintWriter stream)
    {
        System.out.println("in display N region");
        String GeneName = "";
        String UMS = "emitted sequence:  ";
        String regionString = "";
        int ums_start_location = -1;
        int ums_end_location = -1;
        int ums_index = 0;
        int state_index = 1;
        boolean foundStateRegion = false;
        Annotation anno;
        StateInfo si;
        char geneToken;
        do
        {
            anno = viterbi_state_seq.symbolAt(state_index).getAnnotation();
            si = (StateInfo)anno.getProperty(null);
            geneToken = si.geneToken;
            if(geneToken == state_token)
            {
                foundStateRegion = true;
                ums_start_location = ums_index;
            } else
            if(geneToken == 'X')
            {
                state_index++;
            } else
            {
                state_index++;
                ums_index++;
            }
        } while(state_index <= viterbi_state_seq.length() && !foundStateRegion);
        if(!foundStateRegion)
        {
            return null;
        }
        for(; geneToken == state_token; geneToken = si.geneToken)
        {
            UMS = UMS + ums.charAt(ums_index);
            regionString = regionString + ums.charAt(ums_index);
            state_index++;
            ums_index++;
            if(state_index > viterbi_state_seq.length())
            {
                break;
            }
            anno = viterbi_state_seq.symbolAt(state_index).getAnnotation();
            si = (StateInfo)anno.getProperty(null);
        }

        ums_end_location = ums_index - 1;
        anno = viterbi_state_seq.symbolAt(state_index - 1).getAnnotation();
        si = (StateInfo)anno.getProperty(null);
        GeneName = si.geneName + "  input start index:  " + ums_start_location + "  input end index:  " + ums_end_location;
        //System.out.println(GeneName);
        //System.out.println(UMS + "\n");
        //System.out.println("\n");
        stream.println(GeneName);
        stream.println(UMS);
        //stream.println("\n");
        
        RegionInfo tempRegionInfo = new RegionInfo(regionString, ums_start_location, ums_end_location);
        return tempRegionInfo;
    }

    private GeneInfo displayGene(String ums, SymbolList viterbi_state_seq, char state_token, PrintWriter stream)
    {
        String GeneName = "";
        String Nucleotide_Numbers = "";
        String Gene = "germline:   ";
        String geneString = "";
        String UMS = "input:      ";
        String umsString = "";
        String Match = "match:      ";
        int first_nucleotide_number = -1;
        int last_nucleotide_number = -1;
        int ums_index = 0;
        int state_index = 1;
        boolean foundStateRegion = false;
        Annotation anno;
        StateInfo si;
        char geneToken;
        do
        {
            anno = viterbi_state_seq.symbolAt(state_index).getAnnotation();
            si = (StateInfo)anno.getProperty(null);
            geneToken = si.geneToken;
            if(geneToken == state_token)
            {
                foundStateRegion = true;
                first_nucleotide_number = si.getNucleotidePosition();
            } else
            if(geneToken == 'X')
            {
                state_index++;
            } else
            {
                state_index++;
                ums_index++;
            }
        } while(state_index <= viterbi_state_seq.length() && !foundStateRegion);
        if(!foundStateRegion)
        {
            return null;
        }
        for(; geneToken == state_token; geneToken = si.geneToken)
        {
            UMS = UMS + ums.charAt(ums_index);
            umsString = umsString + ums.charAt(ums_index);
            Gene = Gene + si.preEmissionSymbol;
            geneString = geneString + si.preEmissionSymbol;
            if(si.preEmissionSymbol == ums.charAt(ums_index))
            {
                Match = Match + '.';
            } else
            {
                Match = Match + '|';
            }
            state_index++;
            ums_index++;
            if(state_index > viterbi_state_seq.length())
            {
                break;
            }
            anno = viterbi_state_seq.symbolAt(state_index).getAnnotation();
            si = (StateInfo)anno.getProperty(null);
        }

        anno = viterbi_state_seq.symbolAt(state_index - 1).getAnnotation();
        si = (StateInfo)anno.getProperty(null);
        last_nucleotide_number = si.getNucleotidePosition();
        GeneName = si.geneName + "  length of complete gene:  " + si.completeGeneLength + "  start: " + first_nucleotide_number + "  end: " + last_nucleotide_number;
        //System.out.println(GeneName);
        //System.out.println(UMS + "\n");
        //System.out.println(Gene + "\n");
        //System.out.println(Match + "\n");
        //System.out.println("\n");
        
        stream.println(GeneName);
        stream.println(UMS);
        stream.println(Gene);
        stream.println(Match);
        //stream.println("\n");
        int completeGeneLength = si.completeGeneLength;
        GeneInfo tempGeneInfo = new GeneInfo(si.geneName, geneString, umsString, completeGeneLength, first_nucleotide_number, last_nucleotide_number);
        return tempGeneInfo;
    }

    private String removeFastaStyleWhiteSpace(String string)
    {
        String noWhiteSpaceString = "";
        for(int i = 0; i < string.length(); i++)
        {
            char curr = string.charAt(i);
            if(curr != '-')
            {
                noWhiteSpaceString = noWhiteSpaceString + curr;
            }
        }

        return noWhiteSpaceString;
    }

    public void run()
    {
         System.out.println("about to search for the v genes");
        startAlignment();
    }
}
