#/usr/bin/perl -w

#script to run batch iHMMune align
#each iteration of the script performs a given number of alignments (not more than 50 recommended, and must be less than 100)

#define some of the arguments that need to be passed to the iHMMune Align program
$javaCommand = 'java -Xms1512m -Xmx1512m iHMMuneAlign.nonGUI';
$sequenceFilename = 'Barington.fasta';
$vRepertoireName = 'Standard_IGHV_Repertoire.fa';
$dRepertoireName = 'Standard_IGHD_repertoire.fa';
$jRepertoireName = 'j-regions.fa';
$probabilitiesName = 'Current iHMMune-align Probabilities.PH';
$outputFileName = "Barington " . time . " " ; #no file extension necessary
$scoringType = '1'; #can be 1 for HS or 0 for trinucleotide
$outputType = '3';  #1 for excel, 2 for web, 3 for csv
$dAcceptance = '0'; # 5 for minimum of 5 consecutive matches, 8 for minimum of 8 consecutive matches, 0 for dynamic minimum for D misid probs
$alignmentType = '4'; #1 for single sequence in single file, 2 for multiple in multiple files, 4 for mulitple in appended file
$scoringMatrix = '.\\ig_PAM3.txt'; #file name that contains the scoring matrix
$minJLength = '40';
$alignmentModelType = 1; #indicate the alignment type whether full (1), allow partial (2), ighv only (3)
$noOfVGenes = 1; #number of germline IGHV to return

#over ride the minJLength if the alignment model type is set to 2 (partial)
#if ($alignmentModelType == 2) {
#  $minJLength = 0;
#}

#number of sequences to process with each call to iHMMune Align
$seqPerRun = 20;

#sequence to start at
$startIndex = 1;
#sequence to finish at
$endIndex = 10000;

#count the number of sequences in the input file
#open up the sequence file
unless (open (INPUT, $sequenceFilename)) {
  print "Could not open the sequence file: $sequenceFilename.\n";
  exit;
}

#set counter to 0
$inputCount = 0;
while ($currLine = <INPUT>) {
  #if using FASTSA format sequence may be over multiple lines - so just count lines with '>' symbol
  
  if ($currLine =~ m/\>/) {
    #increment the count
    $inputCount++;
  }

  #skip over the line that contains the sequence information
  #$currLine = <INPUT>;
} #end looping through the sequence file


#close the input file cause there is no need for it any longer
close INPUT;

#check that the end index is less than or equal to the number of sequecences in the file
#if not then set end index equal to the number of sequences in the file
if ($endIndex > $inputCount) {
  #if the end index is greater than the number of input sequences
  #set endIndex to the inputcount
  $endIndex = $inputCount;
  
  print "The end index was greater than the number of sequences in the file, so end index set to number of sequence in the file.\n";
}

#check that the startIndex is less than the end index
if ($startIndex > $endIndex) {
  #if the start is greater than finish then swap them around
  $tempEnd = $endIndex;
  $tempStart = $startIndex;

  $endIndex = $tempStart;
  $startIndex = $tempEnd;

  print "The start and end indexes were swapped as the start position must be less than the end position.\n";
}

$processingSeqs = 1;
$prevEndIndex = $startIndex - 1;
while ($processingSeqs == 1) {
  #determine the start index for this run
  $currStart = $prevEndIndex + 1;

  #determine the end index for this run
  $currEnd = (($currStart + $seqPerRun) - 1);

  #check that the current end index is within allowable range
  if ($currEnd >= $endIndex) {
    #set to the endIndex
    $currEnd = $endIndex;
    #change the processingSeqs flag because we are done
    $processingSeqs = 0;
  }
  
  #call iHMMuneAlign
  print "Running command:\n";
  print "$javaCommand \"$sequenceFilename\" \"$vRepertoireName\" \"$dRepertoireName\" \"$jRepertoireName\" \"$probabilitiesName\" \"$outputFileName $currStart-$currEnd\" \"$scoringType\" \"$outputType\" \"$dAcceptance\" \"$alignmentType\" \"$currStart\" \"$currEnd\" \"$scoringMatrix\" \"$minJLength\" \"$alignmentModelType\" \"$noOfVGenes\"\n";
  system "$javaCommand \"$sequenceFilename\" \"$vRepertoireName\" \"$dRepertoireName\" \"$jRepertoireName\" \"$probabilitiesName\" \"$outputFileName $currStart-$currEnd\" \"$scoringType\" \"$outputType\" \"$dAcceptance\" \"$alignmentType\" \"$currStart\" \"$currEnd\" \"$scoringMatrix\" \"$minJLength\" \"$alignmentModelType\" \"$noOfVGenes\"";

  #pause the script so that we can see the error
  #sleep(10);

  #save the current end value to the previous end value variable for use in the next round
  $prevEndIndex = $currEnd;

}#end while processing seqs

#exit the script
exit;