package iHMMuneAlign;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.ArrayList;
import java.util.Vector;
import org.biojava.bio.seq.Sequence;

// Referenced classes of package iHMMuneAlign:
//            GlobalDefines, MutabilityScoreTrinucleotide, MutabilityScoreHotspot, ProbabilityHolder,
//            FastaReader, OutputWindow, AlignmentThread, MutabilityScore

public class nonGUI
    implements GlobalDefines
{
    ArrayList main_sequence_list;
    File V_GENES_FILE;
    File D_GENES_FILE;
    File J_GENES_FILE;
    char html_writer_type;
    int d_gene_acceptance_type;
    MutabilityScoreTrinucleotide mutabilityScoreTrinucleotide;
    MutabilityScoreHotspot mutabilityScoreHotspot;
    MutabilityScore mutabilityScore;
    PrintWriter pw;
    PrintWriter out;

    public nonGUI(String sequenceSetFileName, String vRepFilename, String dRepFilename, String jRepFilename, String probabilityFilename, String htmlFileName, int scoringSystem, int htmlWriterType, int dAcceptanceLevel, byte alignment_type, int startIndex, int endIndex, String scoringMatrix, int minJLength, int alignmentModelType, int numOfVgenes)
    {
        
        System.out.print("alignment model type from the commmand line " + alignmentModelType);

        main_sequence_list = null;

        //names of the repertoire files -> get these from the args list
        V_GENES_FILE = new File(vRepFilename);
        D_GENES_FILE = new File(dRepFilename);
        J_GENES_FILE = new File(jRepFilename);

        mutabilityScoreTrinucleotide = null;
        mutabilityScoreHotspot = null;
        mutabilityScore = null;
        pw = null;

        //determine the output type
        if (htmlWriterType == 1) {
          html_writer_type = '\001'; //excel based output
          //use html extension
          htmlFileName = htmlFileName + ".html";
        } else if (htmlWriterType == 2) {
          html_writer_type = '\002';  //web based output
          //use html extension for output file
          htmlFileName = htmlFileName + ".html";
        } else {
          html_writer_type = '\003';
          //change the output file to the text extension rahter than html
          htmlFileName = htmlFileName + ".txt";
        }

        //read in the sequences from the specified file
        ArrayList sequenceList = loadSequenceSequenceSet(sequenceSetFileName);
        setSequenceList(sequenceList);
        //load the probabilities from the file -> get the file from the args list
        loadProbabilities(probabilityFilename);
        //create the scoring systems
        mutabilityScoreTrinucleotide = new MutabilityScoreTrinucleotide();
        mutabilityScoreHotspot = new MutabilityScoreHotspot();
        
            //stuff from the old action listener function
            if (scoringSystem == 0) {
               mutabilityScore = mutabilityScoreTrinucleotide;    //scoring based on trinucleotides
            }
            else {
               mutabilityScore = mutabilityScoreHotspot;          //scoring based on hotspots
            }

            //select the level of D gene acceptance
            if (dAcceptanceLevel == 5) {
               d_gene_acceptance_type = 2;     //5mer required
            }
            else if (dAcceptanceLevel == 8) {
               d_gene_acceptance_type = 1;    //8mer required
            } 
            else if(dAcceptanceLevel == 0) {   //dynamic setting of the d acceptance
               d_gene_acceptance_type = 3;
            }

            //multiple alignment
            try
            {
                pw = new PrintWriter(new FileWriter("mutations in alignments.txt"));
            }
            catch(Exception ex)
            {
                throw new Error(ex.getMessage());
            }
            
            //selectedIndices is a array of integers that lists all sequences for alignment eg 1,2,3,4,5,6,7,8,9,10 does the first 10 seqs
            int indexStart = startIndex;
            int indexEnd = endIndex;
            int numberOfAlignments = ((indexEnd - indexStart) + 1);
            if (numberOfAlignments > 99) {
              //don't want to be performing more than 100 alignments in one go unless array size is changed for selectedIndicies
              //so set the indexEnd back to something smaller
              indexEnd = indexStart + 99;
              numberOfAlignments = 100;
            }
            int currentIndex = indexStart;
            System.out.println("current index set to " + indexStart);
            int[] selectedIndices = new int[100];  //change to get indices from the args list
            for (int k=0; k < numberOfAlignments; k++) {
              selectedIndices[k] = currentIndex;
              currentIndex++;
              System.out.println("current index in making array = " + currentIndex);
            }

            //change to output this stuff to file rather then to the GUI window
            //OutputWindow tempOutputWindow = new OutputWindow("Multiple Sequence Alignment Result");
            //JTextArea outputArea = tempOutputWindow.getTextArea();

            System.out.println("selected number of sequences = " + selectedIndices.length);
            for(int i = 0; i < numberOfAlignments; i++)
            {
                int currSelectionIndex = selectedIndices[i];
                Sequence currSequence = (Sequence)main_sequence_list.get(currSelectionIndex);
                System.out.println("aligning number i = " + (i + 1));
                alignSequence(currSequence, htmlFileName, alignment_type, d_gene_acceptance_type, probabilityFilename, scoringMatrix, minJLength, alignmentModelType, numOfVgenes);
                System.out.println("finished aligning number i = " + (i + 1));
            }

            if(pw != null)
            {
                pw.close();
            }
            
            //out.close();
            // end of the mulitple aligment code
    }

    private void setSequenceList(ArrayList sequenceList)
    {
        Vector ums_vector = new Vector();
        Vector ums_name_vector = new Vector();
        for(int i = 0; i < sequenceList.size(); i++)
        {
            Sequence tempSeq = (Sequence)sequenceList.get(i);
            String indexed_sequence_name = (i + 1) + "\t: " + tempSeq.getName();
            ums_name_vector.add(indexed_sequence_name);
            ums_vector.add(tempSeq.seqString());
        }

    }


    private void loadProbabilities(String openFileName)
    {
        File openFile = new File(openFileName);
        try
        {
            BufferedReader br = new BufferedReader(new FileReader(openFile));
            String input = "";
            for(String line = br.readLine(); line != null; line = br.readLine())
            {
                input = input + line + " ";
            }

            System.out.println(input);
            ProbabilityHolder newProbabilityHolder = ProbabilityHolder.createFromString(input);
        }
        catch(IOException ioe)
        {
            throw new Error("loadProbsButton failed:  " + ioe.getMessage());
        }
    }

    private ArrayList loadSequenceSequenceSet(String openFileName)
    {
        System.out.println("loadSequenceSequenceSet function called with fileName: " + openFileName);
        File openFile = new File(openFileName);
        ArrayList sequenceList = null;
        FastaReader fr = new FastaReader();
        sequenceList = fr.readFile(openFile);
        if(sequenceList == null)
        {
            System.out.println("Error: Could not read Sequence List FASTA file: " + openFile.getName());
            return null;
        } else
        {
            main_sequence_list = sequenceList;
            return sequenceList;
        }
    }

    private void alignSequence(Sequence sequence, String htmlFileName, byte alignmentType, int dGeneAcceptanceType, String probabilityFilename, String scoringMatrix, int minJLength, int alignmentModelType, int numOfVgenes)
    {
        if(alignmentType != 4)
        {
            String sequenceName = sequence.getName();
            htmlFileName = htmlFileName + " - " + sequenceName;
        }
        //ProbabilityHolder probHolder = ;
        File openProbFile = new File(probabilityFilename);
        try
        {
            BufferedReader br = new BufferedReader(new FileReader(openProbFile));
            String input = "";
            for(String line = br.readLine(); line != null; line = br.readLine())
            {
                input = input + line + " ";
            }

            System.out.println("Input : " + input);
            ProbabilityHolder probHolder = ProbabilityHolder.createFromString(input);
            AlignmentThread alignmentThread = new AlignmentThread(sequence, V_GENES_FILE, D_GENES_FILE, J_GENES_FILE, htmlFileName, probHolder, mutabilityScore, html_writer_type, alignmentType, dGeneAcceptanceType, scoringMatrix, minJLength, alignmentModelType, numOfVgenes);
                alignmentThread.start();
                if(alignmentType == 4)
                {
                    if(!alignmentThread.isAlive());
                    try
                    {
                        alignmentThread.join();
                    }
                    catch(InterruptedException ie)
                    {
                        throw new Error("alignment thread interrupted:  " + ie.getMessage());
                    }
                }

        }
        catch(IOException ioe)
        {
            throw new Error("loadProbsButton failed:  " + ioe.getMessage());
        }

     }

    public static void main(String args[])
    {
        String startupSequencesFileName;
        String vRepFilename;
        String dRepFilename;
        String jRepFilename;
        String probabilityFilename;
        String htmlFilename;
        String txtFilename;
        int scoringSystem;
        int htmlWriterType;
        int dAcceptanceLevel;
        byte alignment_type;
        int startIndex;
        int endIndex;
        String scoringMatrix;
        int minJLength;
        int alignmentModelType;
        int numOfVgenes;

        try {
           startupSequencesFileName = args[0];
           vRepFilename = args[1];
           dRepFilename = args[2];
           jRepFilename = args[3];
           probabilityFilename = args[4];
           htmlFilename = args[5];
           scoringSystem = Integer.parseInt(args[6]);
           htmlWriterType = Integer.parseInt(args[7]);
           dAcceptanceLevel = Integer.parseInt(args[8]);
           alignment_type = Byte.parseByte(args[9]);
           startIndex = Integer.parseInt(args[10]);
           endIndex = Integer.parseInt(args[11]);
           scoringMatrix = args[12];
           minJLength = Integer.parseInt(args[13]);
           alignmentModelType = Integer.parseInt(args[14]);
           numOfVgenes = Integer.parseInt(args[15]);


           //adjust these as the indexing of the list of sequences start at 0 rather than 1
           startIndex = startIndex -1;
           endIndex = endIndex -1;

	} catch (Exception ex) {
	   startupSequencesFileName = "Andrews Sequences All - altered.txt";
           //filenames for the repertoire files
           vRepFilename = "v-regions.fa";
           dRepFilename = "d-regions-no inverts.fa";
           jRepFilename = "j-regions.fa";
           probabilityFilename = "Current iHMMune-align Probabilities.PH";
           htmlFilename = "test outputFile";

           //type of scoring required
           scoringSystem = 1; //0 for trinucleotide, 1 for HS
           //the type of output required -> get from the args list
           htmlWriterType = 1;

           //the level of IGHD gene acceptance
           dAcceptanceLevel = 5; //for fivermer can set to 8 for 8mers, defaults to 8mer if nothing supplied

           //single or multiplefile output
           //byte alignment_type = 2; //for individual file output
           alignment_type = 4;      //for single file output
           
           //some default values for the start and end indicies
           startIndex = 0;
           endIndex = 49;
           
           //the file containing the scoring matrix
           scoringMatrix = "NUC.4.4.txt";
           //minimum IGHJ length required
           minJLength = 30;
           
           //type of alignment - full, partial, ighv only
           alignmentModelType = 1;
           
           //number of vgenes to return alignment for
           numOfVgenes = 5;

	}

        nonGUI myNonGUI = new nonGUI(startupSequencesFileName, vRepFilename, dRepFilename, jRepFilename, probabilityFilename, htmlFilename, scoringSystem, htmlWriterType, dAcceptanceLevel, alignment_type, startIndex, endIndex, scoringMatrix, minJLength, alignmentModelType, numOfVgenes);
        //GUI myGUI = new GUI(startupSequencesFileName);
        //myGUI.show();
    }


}
